#include <iostream>
#include <sstream>
#include <fstream>
#include <vector>
#include <string>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#define numberof(x)	(sizeof(x)/sizeof(*(x)))

using namespace std;

typedef double real;

void printFloat(FILE *file, real f) {
	//cout << t;	//I'd use cout and its Xe+/-Y format, but gnuplot is being stumped by these ...
	fprintf(file, "%.40f", f);
}

struct Problem {
	virtual string name() = 0;
	
	//function and derivatives:
	virtual real x(real t) = 0;
	virtual real dxdt(real x, real t) = 0;
	virtual real ddx_dxdt(real x, real t) = 0;
	virtual real ddx_dtt(real x, real t) = 0;

	//values:
	virtual real x0() = 0;
	virtual real t0() = 0;
	virtual real tn() = 0;
	virtual real dt() = 0;
	
	//implicit integrators need to be defined per-problem
	virtual real implicitBackwardEuler(real x, real t, real dt) {
		/*
		simple cheat implicit integrator -- solve newton assuming the deriv is linear in terms of the x(t+dt) ...
		newton is the best way to go if you end up with a nonlinear / high degree polynomial system  -- which is most likely
		in such a case you can assume that the deriv of the function at the next timestep x wrt the next timestep x is zero (no relation)
		and you get this quick and clever iterative convergence
		if you have any more such value (like the actual relation of f(x(t+dt),t+dt) to x(t+dt)) then you can write a better newton solver 
		
		(x(t+dt) - x(t)) / dt = f(x(t+dt),t+dt)
		x(t+dt) - dt f(x(t+dt),t+dt) - x(t)
		let g(x(t+dt)) = x(t+dt) - dt f(x(t+dt),t+dt) - x(t)
		to do newton root finding, find...
		d/dx(t+dt)[g] = 1 - dt d/dx(t+dt)[f(x(t+dt),t+dt)]
		assume d/d[x(t+dt)] of f(x(t+dt),t+dt) is 0 to get d/dx(t+dt)[g] = 1
		apply to newton root finding iteration: 
		x(t+dt)[k+1] = x(t+dt)[k] - (x(t+dt)[k] - dt f(x(t+dt)[k],t+dt) - x(t))
		x(t+dt)[k+1] = x(t) + dt f(x(t+dt)[k],t+dt)
		*/
		const real convergenceThreshold = 1e-9;
		real xn = x;
		real xp = x;
		for (int iter = 0; iter < 1000; iter++) {
			xn = x + dt * dxdt(xp, dt);
			if (abs(xn - xp) < convergenceThreshold) break; 
			xp = xn;
		}
		return xn;
	}

	virtual real implicitTrapezoid(real x, real t, real dt) {
		//do the newton method by default
		return implicitBackwardEuler(x,t,dt);
	}
};

struct SpiralProblem : public Problem {
	/*
	 exponent * sine wave prediction
	 dx/dt = exp(2. * t) * sin(t)
	 */
	string name() { return "spiral"; }
	real x(real t) { return exp(2. * t) * sin(t); }
	real dxdt(real x, real t) { return x * cos(t); }	//return exp(2. * t) * (2. * sin(t) + cos(t)); }
	real ddx_dxdt(real x, real t) { return cos(t); }
	real ddx_dtt(real x, real t) { return dxdt(x,t) * cos(t) - x * sin(t); } 
	real x0() { return 0.; }
	real t0() { return 0.; }
	real tn() { return 20.; }
	real dt() { return 1.; }
};


struct ExponentProblem : public Problem {
	/*
	 exponent
	 dx/dt = x = exp(t)
	 */
	string name() { return "exp"; }
	real x(real t) { return exp(t); }
	real dxdt(real x, real t) { return x; }
	real ddx_dxdt(real x, real t) { return 1; }
	real ddx_dtt(real x, real t) { return x; }
	real x0() { return 1.; }
	real t0() { return 0.; }
	real tn() { return 100.; }
	real dt() { return .5; }
};

struct ExponentDecayProblem : public Problem {
	static const real C = -15.;
	string name() { 
		ostringstream ss;
		ss << "exp_" << C;
		return ss.str(); 
	}
	real x(real t) { return exp(C * t); }
	real dxdt(real x, real t) { return C * x; }
	real ddx_dxdt(real x, real t) { return C; }
	real ddx_dtt(real x, real t) { return C * C * x; }
	real x0() { return 1.; }
	real t0() { return 0.; }
	real tn() { return 1.; }
	real dt() { return .125; }

	/*
	dx/dt = C * x
	(x(t+dt) - x(t)) / dt = C * x(t+dt)
	x(t+dt)/dt - C * x(t+dt) = x(t) / dt
	(1 - C)/dt * x(t+dt) = x(t) / dt
	x(t+dt) = x(t) / (1 - C)
	*/
	real implicitBackwardEuler(real x, real t, real dt) {
		return x / (1. - C);
	}
	
	/*
	dx/dt = C * x
	(x(t+dt) - x(t)) / dt = (C * x(t) + C * x(t+dt))/2
	x(t+dt)/dt - C * x(t+dt)/2 = x(t)/dt + C * x(t)/2
	x(t+dt)*(1/dt - C/2) = x(t)*(1/dt + C/2)
	x(t+dt) = x(t)*(1/dt + C/2)/(1/dt - C/2)
	x(t+dt) = x(t)*(2 + dt*C)/(2 - dt*C)
	*/
	real implicitTrapezoid(real x, real t, real dt) {
		real Cdt = C * dt;
		return x * (2. + Cdt) / (2. - Cdt);
	}
};

struct SineProblem : public Problem {
	/*
	 sine wave prediction
	 x = sin(t)
	 so dx/dt = cos(t)
	 */
	string name() { return "sine"; }
	real x(real t) { return sin(t); }
	real dxdt(real x, real t) { return cos(t); }
	real ddx_dxdt(real x, real t) { return 0.; }	//worst-case d/dx[dx/dt(t)] = 0, this means B.E.Newton becomes a fixed-point convergence of the derivative
	real ddx_dtt(real x, real t) { return -x; }
	real x0() { return 0.; }
	real t0() { return 0.; }
	real tn() { return 100.; }
	real dt() { return 1; }
};

struct TangentProblem : public Problem {
	/*
	 here's a popular integration problem:
	 y' = tan(y) + 1, y0 = 1, t in [1,1.1}, h = 0.025
	 */
	string name() { return "tangent"; }
	real x(real t) { return -1.; }	//not a number?  don't return anything? don't print anything, that's for sure...
	real dxdt(real x, real t) { return tan(x) + 1.; }
	real ddx_dxdt(real x, real t) { return 0; }
	real ddx_dtt(real x, real t) { return dxdt(x,t) / (cos(x) * cos(x)); }
	real x0() { return 1.; }
	real t0() { return 1.; }
	real tn() { return 1.1; }
	real dt() { return .025; }
};

struct Integrator {
	Problem *problem;
	real x;
	real t;
	real dt;
	
	FILE *valueLog;
	FILE *errorLog;

	//initial x value is here:
	Integrator()
	:	problem(NULL),
		valueLog(NULL),
		errorLog(NULL),
		x(0.),
		t(0.),
		dt(1.)
	{
	}
	
	virtual string logName() = 0;
	
	virtual string valueLogFileName() { return logName() + "_" + problem->name() + ".log"; }
	virtual string errorLogFileName() { return logName() + "_" + problem->name() + "_err.log"; }

	virtual void startLog() {
		string valueLogName = valueLogFileName();
		string errorLogName = errorLogFileName();
		valueLog = fopen(valueLogName.c_str(), "w");
		//old fashioned exception handling
		if (!valueLog) {
			fprintf(stderr, "failed to open log file %s\n", valueLogName.c_str());
			exit(0);
		}

		errorLog = fopen(errorLogName.c_str(), "w");
		if (!valueLog) {
			fprintf(stderr, "failed to open log file %s\n", errorLogName.c_str());
			exit(0);
		}
	}
	
	virtual void stopLog() {
		fclose(valueLog);
		fclose(errorLog);
	}
	
	//typically "t x"
	virtual void logState() {
		printFloat(valueLog, t);
		fprintf(valueLog, "\t");
		printFloat(valueLog, x);
		fprintf(valueLog, "\n");

		printFloat(errorLog, t);
		fprintf(errorLog, "\t");
		printFloat(errorLog, fabs(x - problem->x(t)));
		fprintf(errorLog, "\n");
	}
	
	virtual void run(Problem *problem_) {
		problem = problem_;
		startLog();
		x = problem->x0();
		t = problem->t0();
		dt = problem->dt();
		logState();
		while (t <= problem->tn()) {
			integrate();
			logState();
		}
		
		stopLog();
	}
	
	real dxdt(real x, real t) { return problem->dxdt(x,t); }
	real ddx_dxdt(real x, real t) { return problem->ddx_dxdt(x,t); }
	real ddx_dtt(real x, real t) { return problem->ddx_dtt(x,t); }

	virtual void integrate() = 0;	
};

//explicit Runge-Kutta methods:

struct Euler : public Integrator {
	virtual string logName() { return "euler"; }
	virtual void integrate() {
		x += dt * dxdt(x,t);
		t += dt;
	}
};

/*
  0  |
 1/2 | 1/2
-----+----------------
     |  0   1
 */
struct Midpoint : public Integrator {
	virtual string logName() { return "midpoint"; }
	virtual void integrate() {
		real k = dt * dxdt(x,t);
		x += dt * dxdt(x + .5 * k, t + .5 * dt);
		t += dt;
	}
};

struct Heun : public Integrator {
	virtual string logName() { return "heun"; }
	virtual void integrate() {
		real dxdtAtX = dxdt(x,t);
		real xTilde = x + dt * dxdtAtX;
		x += .5 * dt * (dxdtAtX + dxdt(xTilde, t + dt));
		t += dt;
	}
};

//alpha = .5 <=> midpoint method
//alpha = 1 <=> Heun's method
struct RK2Alpha : public Integrator {
	real alpha;
	RK2Alpha(real alpha_) : alpha(alpha_) {}
	virtual string logName() {
		ostringstream s;
		s << "rk2a_" << alpha;
		return s.str();
	}
	virtual void integrate() {
		real dxdtAtX = dxdt(x,t);
		real k = dt * dxdtAtX;
		real frac = 1. / (2. * alpha);
		x += dt * ((1. - frac) * dxdtAtX + frac * dxdt(x + alpha * k, t + alpha * dt));
		t += dt;
	}
};

/*
  0  |
 1/2 | 1/2
 1/2 |  0  1/2
  1  |  0   0   1
-----+----------------
     | 1/6 1/3 1/3 1/6
 */
struct RungeKutta4 : public Integrator {
	virtual string logName() { return "rk4"; }
	virtual void integrate() {
		real k1 = dt * dxdt(x,t);
		real k2 = dt * dxdt(x + .5 * k1, t + .5 * dt);
		real k3 = dt * dxdt(x + .5 * k2, t + .5 * dt);
		real k4 = dt * dxdt(x + k3, t + dt);
		x += (k1 + 2. * k2 + 2. * k3 + k4) / 6.;
		t += dt;
	}
};

/*
 0
 1/4	1/4
 3/8	3/32	9/32
 12/13	1932/2197	−7200/2197	7296/2197
 1		439/216	−8	3680/513	-845/4104
 1/2	−8/27	2	−3544/2565	1859/4104	−11/40	
 -------------------------------------------------
 16/135	0	6656/12825	28561/56430	−9/50	2/55
 25/216	0	1408/2565	2197/4104	−1/5	0
 
 */
struct RungeKuttaFehlberg45 : public Integrator {
	real errorThreshold;
	RungeKuttaFehlberg45(real errorThreshold_) : errorThreshold(errorThreshold_) {}

	virtual string logName() {
		ostringstream s;
		s << "rk45_" << errorThreshold;
		return s.str();
	}
	
	//this dt overshadows the class dt ... i should rename it or something
	virtual bool integrateAttempt(real dt) {
		
		real k1 = dt * dxdt(x,t);
		real k2 = dt * dxdt(x + k1 / 4., t + dt / 4.);
		real k3 = dt * dxdt(x + k1 * 3./32. + k2 * 9./32., t + dt * 3./8.);
		real k4 = dt * dxdt(x + k1 * 1932./2197. - k2 * 7200./2197. + k3 * 7296./2197., t + dt * 12./13.);
		real k5 = dt * dxdt(x + k1 * 439./216. - k2 * 8. + k3 * 3680./513. - k4 * 845./4104., t + dt);
		real k6 = dt * dxdt(x - k1 * 8./27. + k2 * 2. - k3 * 3544./2565. + k4 * 1859./4104. - k5 * 11./40., t + dt / 2.);
		//high order step
		real xHi = x + k1 * 16./135. + k3 * 6656./12825. + k4 * 28561./56430. - k5 * 9./50. + k6 * 2./55.;
		//low order step
		real xLo = x + k1 * 25./216. + k3 * 1408./2565. + k4 * 2197./4104. - k5 / 5.;
		//and the error between:
		real error = xHi - xLo;
		//now if the error is too big then reduce the step size.
		
		if (fabs(error) >= fabs(errorThreshold)) return false;

		x = xHi;
		
		return true;
	}
	
	virtual void integrate() {
		real dtAttempt = dt;
		while (!integrateAttempt(dtAttempt)) {
			dtAttempt *= .5;
		}
		t += dtAttempt;
	}
};

struct AdamsBashforth : public Integrator {
	
	struct state {
		state() : x(0.), t(0.), dxdt(0.) {}
		state(real x_, real t_, real dxdt_) : x(x_), t(t_), dxdt(dxdt_) {}
		state(const state &s) : x(s.x), t(s.t), dxdt(s.dxdt) {}
		real x;
		real t;
		real dxdt;
	};
	
	//for AB 5, store the last 4 states in addition to the 5th
	vector<state> last;
	int lastSize;
	
	AdamsBashforth() : lastSize(0) {
	}
	
	virtual string logName() { return "adams-bashforth"; }
	
	real dxdt(real x, real t) { return Integrator::dxdt(x, t); }
	real dxdt(const state &s) { return Integrator::dxdt(s.x, s.t); }
	
	virtual void integrate() {
		//add our last state to the stack
		//mind you last[n-1] == state(x, t) so for the one before x,t look to last[n-2]
		last.push_back(state(x,t,dxdt(x,t)));
		
		//handles up to AB5
		int n = last.size();
		real xPred;
		switch (n) {
			case 1:
				xPred = x + dt * last[n-1].dxdt;
				x += dt * dxdt(xPred, t + dt);
				break;
			case 2:
				xPred = x + dt * (3./2. * last[n-1].dxdt - 1./2. * last[n-2].dxdt);
				x += dt * (1./2. * dxdt(xPred, t + dt) + 1./2. * last[n-1].dxdt);
				break;
			case 3:
				xPred = x + dt * (23./12. * last[n-1].dxdt - 4./3. * last[n-2].dxdt + 5./12. * last[n-3].dxdt);
				x += dt * (5./12. * dxdt(xPred, t + dt) + 2./3. * last[n-1].dxdt - 1./12. * last[n-2].dxdt);
				break;
			case 4:
				xPred = x + dt * (55./24. * last[n-1].dxdt - 59./24. * last[n-2].dxdt + 37./24. * last[n-3].dxdt - 3./8. * last[n-4].dxdt);
				x += dt * (3./8. * dxdt(xPred, t + dt) + 19./24. * last[n-1].dxdt - 5./24. * last[n-2].dxdt + 1./24. * last[n-3].dxdt);
				break;
			case 5:
				xPred = x + dt * (1901./720. * last[n-1].dxdt - 1387./360. * last[n-2].dxdt + 109./30. * last[n-3].dxdt - 637./360. * last[n-4].dxdt + 251./720. * last[n-5].dxdt);
				x += dt * (251./720. * dxdt(xPred, t + dt) + 323./360. * last[n-1].dxdt - 11./30. * last[n-2].dxdt + 53./360. * last[n-3].dxdt - 19./720. * last[n-4].dxdt);
				break;
			case 6:
				xPred = x + dt * (4277./1440. * last[n-1].dxdt - 2641./480. * last[n-2].dxdt + 4991./720. * last[n-3].dxdt - 3659./720. * last[n-4].dxdt + 959./480. * last[n-5].dxdt - 95./288. * last[n-6].dxdt);
				x += dt * (95./288. * dxdt(xPred, t + dt) + 1427./1440. * last[n-1].dxdt - 133./240. * last[n-2].dxdt + 241./720. * last[n-3].dxdt - 173./1440. * last[n-4].dxdt + 3./160. * last[n-5].dxdt);
				break;
			case 7:
				xPred = x + dt * (198721./60480. * last[n-1].dxdt - 18637./2520. * last[n-2].dxdt + 235183./20160. * last[n-3].dxdt - 10754./945. * last[n-4].dxdt + 135713./20160. * last[n-5].dxdt - 5603./2520. * last[n-6].dxdt * 19087./60480. * last[n-7].dxdt);
				x += dt * (19087./60480. * dxdt(xPred, t + dt) + 2713./2520. * last[n-1].dxdt - 15487./20160. * last[n-2].dxdt + 586./945. * last[n-3].dxdt - 6737./20160. * last[n-4].dxdt + 263./2520. * last[n-5].dxdt - 863./60480. * last[n-6].dxdt);
				break;
			case 8:
			default:
				xPred = x + dt * (16083./4480. * last[n-1].dxdt - 1152169./120960. * last[n-2].dxdt + 242653./13440. * last[n-3].dxdt - 296053./13440. * last[n-4].dxdt + 2102243./120960. * last[n-5].dxdt - 115747./13440. * last[n-6].dxdt + 32863./13440. * last[n-7].dxdt - 5257./17280. * last[n-8].dxdt);
				x += dt * (5257./17280. * dxdt(xPred, t + dt) + 139849./120960. * last[n-1].dxdt - 4511./4480. * last[n-2].dxdt + 123133./120960. * last[n-3].dxdt - 88547./120960. * last[n-4].dxdt + 1537./4480. * last[n-5].dxdt - 11351./120960. * last[n-6].dxdt + 275./24192. * last[n-7].dxdt);
				break;
		}
		
		
		t += dt;
		
	}
};

/*
TODO
the method is solved implicitly with its second derivative of the next timestep 
x(t+dt) = x(t) + dt * dx/dt(x,t) + (.5-beta) dt^2 d2x/dt2(x,t) + beta dt^2 d2x/dt2(x(t+dt),t+dt)
*/
struct NewmarkBeta : public Integrator {
	static const real beta = .25;
	virtual string logName() { return "newmark-beta"; }
	virtual void integrate() {
		//x += dt * dxdt(x,t) + (.5 - beta) * ...; 
	}
};

/*
the method is solved implicitly with its second derivative of the next timestep
x'(t+dt) = x(t) + dt v(t) + dt^2/6 (4 d^2x/dt^2(x(t),t) - d^2x/dt^2(x(t-dt), t-dt))
loop:
	x(t+dt) = x(t) + dt v(t) + dt^2/6 (d^2x/dt^2(x'(t+dt),t+dt) + 2 d^2x/dt^2(x(t),t))
	v(t+dt) = (x'(t+dt) - x(t)/dt + dt/6 (2d^2x/dt^2(x(t+dt),t+dt) + d^2x/dt^2(x(t),t))
*/
struct Beeman : public Integrator {
	real v, xp, vp;
	static const int maxiter = 1000;
	real errorThreshold;
	
	Beeman(real errorThreshold_) : Integrator(), v(0.), xp(0.), vp(0.), errorThreshold(errorThreshold_) {}
	
	virtual string logName() { 
		ostringstream s;
		s << "beeman_" << errorThreshold;
		return s.str();
	}
	
	virtual void run(Problem *problem_) {
		v = problem_->dxdt(problem_->x0(), problem_->t0());
		xp = x;
		vp = v;
		Integrator::run(problem_);
	}
	
	virtual void integrate() {
		real aprev = ddx_dtt(xp, t-dt);
		real a = ddx_dtt(x,t);
		
		real xcorr, vcorr;
		int iter = 0;
		for (; iter < maxiter; iter++) {
			real xpred = x + dt * (v + dt / 6. * (4. * a - aprev));
			real apred = ddx_dtt(xpred, t+dt);
			xcorr = x + dt * (v + dt/6. * (4. * apred - a));
			vcorr = (xcorr - x)/dt + dt/6. * (2. * apred + 5. * a); 
			real xerr = fabs(xpred - xcorr);
			if (xerr < errorThreshold) break;
		}
		if (iter == maxiter) {
			cerr << logName() << " did not converge" << endl;
		}
		
		xp = x;
		vp = v;
		x = xcorr;
		v = vcorr;
	}
};

/*
integral a to b f(x) dx = (b - a)/2 sum i=1,5 w_i f( (b-a)/2 x_i + (a+b)/2 )
n	x							w
1	0							2
2	+/-sqrt(3)/3)				1
3	0							8/9
4	+/-sqrt(3-2sqrt(6/5))/7		(18+sqrt(30))/36
	+/-sqrt(3+2sqrt(6/5))/7		(18-sqrt(30))/36
5	0	128/225
	+/-1/3 sqrt(5-2sqrt(10/7))	(322+13sqrt(70))/900
	+/-1/3 sqrt(5+2sqrt(10/7))	(322-13sqrt(70))/900

*/
struct GaussLagrangeQuadrature5 : public Integrator {

};

//implicit methods

struct BackwardEuler_Implicit : public Integrator {
	virtual string logName() { return "backward-euler-implicit"; }
	virtual void integrate() {
		x = problem->implicitBackwardEuler(x,t,dt);
		t += dt;
	}
};

struct BackwardEuler_Newton : public Integrator {
	real errorThreshold;
	BackwardEuler_Newton(real errorThreshold_) : errorThreshold(errorThreshold_) {}
	
	virtual string logName() {
		ostringstream s;
		s << "backward-euler-newton-" << errorThreshold;
		return s.str();
	}
	
	/*
	x(t+dt) = x(t) + dt * df/dt(t+dt)
	x(t+dt) - dt * dx/dt(t+dt) - x(t) = 0
	F = 0 for F = l.h.s.
	represent F wrt x(t+dt), consider x(t) to be constant,
	so dF/dx = 1 - dt * d/dx[dx/dt(t+dt)]
	apply Newton root-finding, starting at x(t)...
	x[0] = x
	x[n+1] = x[n] - F(x[n]) / dF/dx(x[n])
	*/
	virtual void integrate() {
		real xt = x;	//x(t)
		real xn = x;	//Newton-x
		const int maxiter = 100;
		int iter = 0;
		for (; iter < maxiter; iter++) {
			real F = xn - dt * dxdt(xn,t+dt) - xt;
			real dFdx = 1 - dt * ddx_dxdt(xn,t+dt);
			real delta = F / dFdx;
			xn -= delta;
			if (fabs(delta) < errorThreshold) break;
		}
		if (iter == maxiter) {
			cerr << logName() << " did not converge" << endl;
		} else {
			cerr << logName() << " converged after " << iter << " iterations" << endl;
		}
		x = xn;
		t += dt;
	}
};

struct Trapezoid_Implicit : public Integrator {
	virtual string logName() { return "trapezoid-implicit"; }
	virtual void integrate() {
		x = problem->implicitTrapezoid(x,t,dt);
		t += dt;
	}
};

struct Trapezoid_Newton : public Integrator {
	real errorThreshold;
	Trapezoid_Newton(real errorThreshold_) : errorThreshold(errorThreshold_) {}

	virtual string logName() {
		ostringstream s;
		s << "trapezoid-newton-" << errorThreshold;
		return s.str();
	}

	virtual void integrate() {
		real xt = x;
		real dxdt_t = dxdt(x,t);
		real xn = x;
		const int maxiter = 100;
		int iter = 0;
		for (; iter < maxiter; iter++) {
			real F = xn - .5 * dt * dxdt(xn,t+dt) - xt - .5 * dt * dxdt_t;
			real dFdx = 1 - .5 * dt * ddx_dxdt(xn,t+dt);
			real delta = F / dFdx;
			xn -= delta;
			if (fabs(delta) < errorThreshold) break;
		}
		if (iter == maxiter) {
			cerr << logName() << " did not converge" << endl;
		} else {
			cerr << logName() << " converged after " << iter << " iterations" << endl;
		}
		x = xn;
		t += dt;
	}
};

//just evaluate the answer
struct TargetForIntegration : public Integrator {
	virtual string logName() { return "target"; }
	virtual void integrate() {
		t += dt;
		x = problem->x(t);
	}
};

struct Test {
	Problem *problem;
	
	Test(Problem *problem_)
	: problem(problem_) 
	{ }

	void run() {
		Integrator *integrators[] = {
			//forward ...
			new TargetForIntegration(),
			new Euler(),
			new Heun(),
			new Midpoint(),
			new RungeKutta4(),
			new RungeKuttaFehlberg45(1.),	//any error goes
			new RungeKuttaFehlberg45(1e-4),
			new RungeKuttaFehlberg45(1e-9),
			new AdamsBashforth(),
			//new Beeman(1.),
			//new Beeman(1e-4),
			//new Beeman(1e-9),
			//newton...
			new BackwardEuler_Newton(1e-50),
			new Trapezoid_Newton(1e-50),	//does better than BE Newton, but still not near RK45 without a corrector
			//implicit ...
			new BackwardEuler_Implicit(),
			new Trapezoid_Implicit(),
		};
		
		for (int j = 0; j < numberof(integrators); j++) {
			integrators[j]->run(problem);
		}

		//now for the gnuplot script ...
		string plotScriptFilename = "plot-" + problem->name() + ".script";
		ofstream f(plotScriptFilename.c_str());
		f << "set style data lines" << endl;
		f << "set term png size 1024,768" << endl;
		ostringstream plotline;
		ostringstream ploterrline;
		plotline << "plot ";
		ploterrline << "plot ";
		const char *comma = "", *errcomma = "";
		for (int j = 0; j < numberof(integrators); j++) {
			Integrator *integrator = integrators[j];
			
			plotline << comma << "\"" << integrator->valueLogFileName() << "\" title \'" << integrator->logName() << "\'";
			comma = ", ";
			if (integrator->logName() != "target") {
				ploterrline << errcomma << "\"" << integrator->errorLogFileName() << "\" title \'" << integrator->logName() << "\'";
				errcomma = ", ";
			}
		}
		f << "set output \"plot-" << problem->name() << ".png\"" << endl; 
		f << "set title \"integration methods: " << problem->name() << "\"" << endl;
		f << plotline.str() << endl;
		f << "set output \"plot-" << problem->name() << "-err.png\"" << endl; 
		f << "set log y" << endl;
		f << "set title \"integration error: " << problem->name() << "\"" << endl;
		f << ploterrline.str() << endl;
		f << "unset log y" << endl;
		f.close();

		int systemRes = system( (string("gnuplot ") + plotScriptFilename).c_str() );
		if (systemRes) {
			cerr << "gnuplot failed with " << systemRes << endl;
		}

		for (int j = 0; j < numberof(integrators); j++) {
			delete integrators[j];
		}
	}
};

int main() {
	{ SpiralProblem problem; Test(&problem).run(); }
	{ SineProblem problem; Test(&problem).run(); }
	{ TangentProblem problem; Test(&problem).run(); }
	{ ExponentProblem problem; Test(&problem).run(); }
	{ ExponentDecayProblem problem; Test(&problem).run(); }
}
