/*
problems:
	diffusion: du/dt = c d^2u/dx^2
	propagation: du/dt = c du/dx
	advection: du/dt = u du/dx
	
	wave equation: d^u/dt^2 = c^2 d^2u/dx^2	<- R2 1D problems like this introduce many-banded matrices, which are a bit more complex to solve
	extending the system from 1D to 2D also introduces the same problem
	soooo in all cases
	BLOCK TRIDIAGONAL SOLVER

methods:
	forward-time centered-space
	crank-nicolson
	
backend: blitz? or at least lapack.
gui: gnuplot sucks for 3D graphs.  opengl would allow an easy way to extend this to 2D problems as well.
*/
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <map>
#include <math.h>
#include "defs/defs.h"

using namespace std;

//default: just << it
template<typename T> void printValue(ostream &o, T f) { o << f; }

//special case for reals: scientific notation
template<> void printValue(ostream &o, double f) {
	char buf[256];
	sprintf(buf, "%.40e", f);
	o << buf;
}


struct IntegratorIF;

/*
holds static information ...
and temporary data structures ...
*/
struct Problem {
	typedef double real;
	
	//a mess, I know:

	//save integrator:
	IntegratorIF *integrator;
	
	//init Thomas algorithm stuff
	int N;
	real dx;
	real *a, *b, *c, *v;

	//needs vtable access, and parent ctors don't seem to have that
	void init(int N_) {
		N = N_;
		a = new real[N];
		b = new real[N];
		c = new real[N];
		v = new real[N];
		dx = (xmax() - xmin()) / (real)N;
	}
	
	virtual real u0(real x) = 0;
	virtual real t0() = 0;
	virtual real tn() = 0;
	virtual real dt() = 0;
	virtual real xmin() = 0;
	virtual real xmax() = 0;
	
	//when you get into 1+D integration, problems are becoming more tightly coupled with their solvers
	typedef void (Problem::*integrateFunc)(real *un, real *u, real dt);
	
	virtual void integrateForwardTimeCenteredSpace(real *un, real *u, real dt) = 0;
	virtual void integrateLax(real *un, real *u, real dt) = 0;
	virtual void integrateLaxWendroff(real *un, real *u, real dt) = 0;
	virtual void integrateBackwardTimeCenteredSpace(real *un, real *u, real dt) = 0;
	virtual void integrateCrankNicolson(real *un, real *u, real dt) = 0;
	virtual void integrateTarget(real *un, real *u, real dt) = 0;

	//build integrator with appropriate state data. don't call before calling Problem::init()
	virtual IntegratorIF *createIntegrator() = 0;

	/*
	Thomas algorithm <=> solve tridiagonal matrixes
	'v' is the multiplying vector, 'x' is the solution <=> x = A^-1 * v
	destroys 'v' in the process.
	
	Note that this is made for reals.
	if your state data becomes matrices then ... well, this becomes complicated.
	*/
	void solveTridiagonal(real *x, real *v, int N) {
		for (int i = 1; i < N; i++) {
			real m = a[i] / b[i-1];
			b[i] -= m * c[i-1];
			v[i] -= m * v[i-1];
		}
		x[N-1] = v[N-1] / b[N-1];
		for (int i = N-2; i >= 0; i--) {
			x[i] = (v[i] - c[i] * x[i+1]) / b[i];
		}
	}
};

struct IntegratorIF {
	virtual void run(Problem *problem_, const string &name, Problem::integrateFunc func) = 0;
	double *x;
	double t;
};

template<typename REAL>
struct Integrator : public IntegratorIF {
	typedef REAL real;
	
	Problem *problem;
	int N;
	real *u, *un;
	real dt;
	
	ofstream valueLog;
	
	Integrator(int N_)
	:	problem(NULL),
		N(N_),
		u(new real[N_]),
		un(new real[N_]),
		dt(1)
	{
		x = new real[N_];
		t = 0;
	}
	
	virtual void startLog(const string &name) {
		string valueLogName = name + ".log";
		valueLog.open(valueLogName.c_str());
		//old fashioned exception handling
		if (!valueLog) {
			cerr << "failed to open log file " << valueLogName << endl;
			exit(0);
		}
	}
	
	virtual void stopLog() {
		valueLog.close();
	}
	
	//typically "t x"
	virtual void logState() {
		for (int i = 0; i < N; i++) {
			printValue(valueLog, t);
			valueLog << "\t";
			printValue(valueLog, x[i]);
			valueLog << "\t";
			printValue(valueLog, u[i]);
			valueLog << endl;
		}
		valueLog << endl;
	}
	
	real xForIndex(int i) {
		real xmin = problem->xmin();
		real xmax = problem->xmax();
		return (real)(i+.5) / (real)N * (xmax - xmin) + xmin;
	}
	
	virtual void run(Problem *problem_, const string &name, Problem::integrateFunc func) {
		startLog(name);
		problem = problem_;

		//init u's and x's
		t = problem->t0();
		for (int i = 0; i < N; i++) {
			x[i] = xForIndex(i);
			u[i] = problem->u0(x[i]);
		}
		
		dt = problem->dt();
		logState();
		while (t <= problem->tn()) {
			(problem->*func)(un, u, dt);
			t += dt;
			memcpy(u, un, sizeof(real) * N);
			logState();
		}
		stopLog();
	}
};


/*
du/dt = C d^2u/dx^2
*/
struct HeatDiffuseProblem : public Problem {
	typedef Problem::real real;
	virtual IntegratorIF *createIntegrator() { return integrator = new Integrator<real>(N); }

	static const real C = .5;
	virtual real u0(real x) { return exp(-x*x/.05); }
	virtual real t0() { return 0.; }
	virtual real tn() { return 10.; }
	virtual real dt() { return .1; }
	virtual real xmin() { return -1.; }
	virtual real xmax() { return 1.; }
	
	/*
	(u(x,t+dt) - u(x,t)) / dt = C (u(x+dx,t) - 2 * u(x,t) + u(x-dx,t))/dx
	u(x,t+dt)/dt = u(x,t) + C*dt/dx (u(x+dx,t) - 2 * u(x,t) + u(x-dx,t))
	u(x,t+dt)/dt = u(x,t) + C*dt/dx * u(x+dx,t) - 2 * C*dt/dx * u(x,t) + C*dt/dx * u(x-dx,t)
	u(x,t+dt)/dt = C*dt/dx * u(x+dx,t) + (1 - 2 * C*dt/dx) * u(x,t) + C*dt/dx * u(x-dx,t)
	*/
	virtual void integrateForwardTimeCenteredSpace(real *un, real *u, real dt) {
		real mu = C * dt / dx;
		for (int i = 1; i < N-1; i++) {
			un[i] = mu * u[i+1] + (1. - 2. * mu) * u[i] + mu * u[i-1];
		}
		un[0] = un[1];
		un[N-1] = un[N-2];
	}

	/*
	(u(x,t+dt) - u(x,t)) / dt = C (u(x+dx,t) - 2 * u(x,t) + u(x-dx,t))/dx
	u(x,t+dt)/dt = u(x,t) + C*dt/dx (u(x+dx,t) - 2 * u(x,t) + u(x-dx,t))
	... substitute u(x,t) for (u(x+dx,t) + u(x-dx,t))/2 ...
	u(x,t+dt)/dt = 1/2*(u(x+dx,t) + u(x-dx,t)) + C*dt/dx (u(x+dx,t) - 2 * u(x,t) + u(x-dx,t))
	u(x,t+dt)/dt = 1/2*u(x+dx,t) + 1/2*u(x-dx,t) + C*dt/dx * u(x+dx,t) - 2 * C*dt/dx * u(x,t) + C*dt/dx * u(x-dx,t)
	u(x,t+dt)/dt = (1/2 + C*dt/dx) * u(x+dx,t) - 2 * C*dt/dx * u(x,t) + (1/2 + C*dt/dx) * u(x-dx,t)
	*/
	virtual void integrateLax(real *un, real *u, real dt) {
		real mu = C * dt / dx;
		for (int i = 1; i < N-1; i++) {
			un[i] = (.5 + mu) * u[i+1] - 2. * mu * u[i] + (.5 + mu) * u[i-1];
		}
		un[0] = un[1];
		un[N-1] = un[N-2];
	}

	virtual void integrateLaxWendroff(real *un, real *u, real dt) {
		//TODO
	}

	/*
	(u(x,t+dt) - u(x,t)) / dt = C (u(x+dx,t+dt) - 2 * u(x,t+dt) + u(x-dx,t+dt))/dx
	u(x,t+dt)/dt - C/dx * u(x+dx,t+dt) + 2*C/dx * u(x,t+dt) - C/dx * u(x-dx,t+dt) = u(x,t)/dt
	 -C*dt/dx * u(x+dx,t+dt) + (1 + 2*dt*C/dx) * u(x,t+dt) - C*dt/dx * u(x-dx,t+dt) = u(x,t)
	*/
	virtual void integrateBackwardTimeCenteredSpace(real *un, real *u, real dt) {
		real mu = C * dt / dx;
		for (int i = 0; i < N; i++) {
			a[i] = -mu;
			b[i] = 1. + 2. * mu;
			c[i] = -mu;
		}
		
		solveTridiagonal(un,u,N);
	}

	/*
	(u(x,t+dt) - u(x,t)) / dt = (C (u(x+dx,t) - 2 * u(x,t) + u(x-dx,t))/dx + C (u(x+dx,t+dt) - 2 * u(x,t+dt) + u(x-dx,t+dt))/dx)/2
	u(x,t+dt)/dt - C/(2*dx) * u(x+dx,t+dt) + C/dx * u(x,t+dt) - C/(2*dx) * u(x-dx,t+dt) = u(x,t)/dt + C/(2*dx) * u(x+dx,t) - C/dx * u(x,t) + C/(2*dx) * u(x-dx,t)
	 -C*dt/(2*dx) * u(x+dx,t+dt) + (1 + C*dt/dx) * u(x,t+dt) - C*dt/(2*dx) * u(x-dx,t+dt) = C*dt/(2*dx) * u(x+dx,t) + (1 - C*dt/dx) * u(x,t) + C*dt/(2*dx) * u(x-dx,t)
	*/
	virtual void integrateCrankNicolson(real *un, real *u, real dt) {
		real mu = C * dt / dx;
		for (int i = 0; i < N; i++) {
			a[i] = -.5 * mu;
			b[i] = 1. + mu;
			c[i] = -.5 * mu;
			v[i] = (1. - mu) * u[i];
			if (i > 0) v[i] += .5 * mu * u[i-1];
			if (i < N-1) v[i] += .5 * mu * u[i+1];
		}
		
		solveTridiagonal(un,v,N);
	}
	
	/*
	Method-of-Lines:
	turn the problem into du/dt = S * u
	then do iterative RK on it
	... note that S * u means S = A^-1 * B , for A the tridiagonal matrix we're solving and B another tridiagonal system on the rhs
	... so S(x) would be (1) apply tridiagonal B to x then (2) tridiagonal inverse solve against A
	*/
};

/*
du/dt = c du/dx
//http://www.physics.udel.edu/~jim/Advection/advection.htm


let u(x,t) = v(x+ct) for any C-inf v:x->x
dv/dx = v'
dv/dt = cv'
so there's your solution
*/
struct PropagationProblem : public Problem {
	typedef Problem::real real;
	virtual IntegratorIF *createIntegrator() { return integrator = new Integrator<real>(N); }

	static const real C = -1.;
	virtual real u0(real x) { return exp(-(x-.5)*(x-.5)/.01); }
	virtual real t0() { return 0.; }
	virtual real tn() { return 1.; }
	virtual real dt() { return .01; }
	virtual real xmin() { return 0.; }
	virtual real xmax() { return 1.; }
	
	/*
	du/dt = C*du/dx
	(u(x,d+dt) - u(x,t)) / dt = C * (u(x+dx,t) - u(x-dx,t)) / (2 * dx)
	u(x,d+dt)/dt =  u(x,t) + C * dt / (2 * dx) (u(x+dx,t) - u(x-dx,t))
	*/
	virtual void integrateForwardTimeCenteredSpace(real *un, real *u, real dt) {
		real mu = C * dt / dx;
		for (int i = 1; i < N-1; i++) {
			un[i] = u[i] + .5 * mu * (u[i+1] - u[i-1]);
		}
		un[0] = un[1];
		un[N-1] = un[N-2];
	}
	
	/*
	du/dt = C*du/dx
	(u(x,t+dt) - u(x,t)) / dt = C * (u(x+dx,t) - u(x-dx,t)) / (2 * dx)
	u(x,t+dt) = u(x,t) + C*dt/(2*dx) * (u(x+dx,t) - u(x-dx,t))
	... and this is where we substitute u(x,t) for 1/2(u(x-dx,t) + u(x+dx,t))
	u(x,t+dt) = 1/2*(u(x+dx,t) + u(x-dx,t)) + C*dt/(2*dx) * (u(x+dx,t) - u(x-dx,t))	
	u(x,t+dt) = 1/2*u(x+dx,t) + 1/2*u(x-dx,t) + C*dt/(2*dx) * u(x+dx,t) - C*dt/(2*dx) * u(x-dx,t)
	u(x,t+dt) = (1/2 + C*dt/(2*dx)) * u(x+dx,t) + (1/2 - C*dt/(2*dx)) * u(x-dx,t)
	*/
	virtual void integrateLax(real *un, real *u, real dt) {
		real mu = C * dt / dx;
		for (int i = 1; i < N-1; i++) {
			un[i] = (.5 + .5 * mu) * u[i+1] +  (.5 - .5 * mu) * u[i-1];
		}
		un[0] = un[1];
		un[N-1] = un[N-2];
	}
	
	/*
	u(x,t+dt) = u(x,t) + 1/2 C*dt/dx (u(x+dt,t) - u(x-dx,t)) + 1/2 (C*dt/dx)^2 (u(x+dx,t) - 2 u(x,t) + u(x-dx,t))
	*/
	virtual void integrateLaxWendroff(real *un, real *u, real dt) {
		real mu = C * dt / dx;
		for (int i = 1; i < N-1; i++) {
			un[i] = u[i] + .5 * mu * (u[i+1] - u[i-1]) + .5 * mu * mu * (u[i+1] - 2. * u[i] + u[i-1]);
		}
		un[0] = un[1];
		un[N-1] = un[N-2];
	}
			
	/*
	du/dt = C*du/dx
	(u(x,d+dt) - u(x,t)) / dt = C * (u(x+dx,t+dt) - u(x-dx,t+dt)) / (2 * dx)
	u(x,d+dt) - C*dt/(2*dx) * (u(x+dx,t+dt) - u(x-dx,t+dt)) = u(x,t)
	C*dt/(2*dx) * u(x-dx,t+dt) + u(x,d+dt) - C*dt/(2*dx) * u(x+dx,t+dt) = u(x,t)
	*/
	virtual void integrateBackwardTimeCenteredSpace(real *un, real *u, real dt) {
		real mu = C * dt / dx;
		for (int i = 1; i < N-1; i++) {
			a[i-1] = .5 * mu;
			b[i-1] = 1.;
			c[i-1] = -.5 * mu;
		}
		solveTridiagonal(un+1,u+1,N-2);
		un[0] = un[1];
		un[N-1] = un[N-2];
	}

	
	/*
	du/dt = C*du/dx
	(u(x,d+dt) - u(x,t)) / dt = (C * (u(x+dx,t) - u(x-dx,t)) / (2 * dx) + C * (u(x+dx,t+dt) - u(x-dx,t+dt)) / (2 * dx))/2
	C*dt/(4*dx) * u(x-dx,t+dt) + u(x,d+dt) - C*dt/(4*dx) * u(x+dx,t+dt) = -C*dt/(4*dx) * u(x-dx,t) + u(x,t) + C*dt/(4*dx) * u(x+dx,t)
	*/
	virtual void integrateCrankNicolson(real *un, real *u, real dt) {
		real mu = C * dt / dx;
		for (int i = 1; i < N-1; i++) {
			a[i-1] = .25 * mu;
			b[i-1] = 1.;
			c[i-1] = -.25 * mu;
			v[i] = -.25 * mu * u[i-1] + u[i] + .25 * mu * u[i+1];
		}
		solveTridiagonal(un+1,v+1,N-2);
		//how to make the reflection go away ... 
		un[0] = un[1];
		un[N-1] = un[N-2];
	}
	
	virtual void integrateTarget(real *un, real *u, real dt) {
		for (int i = 0; i < N; i++) {
			un[i] = u0(integrator->x[i] - integrator->t);
		}
	}
};

/*
du/dt = u du/dx = d/dx[1/2 u^2]
*/
struct AdvectionProblem : public Problem {
	typedef Problem::real real;
	virtual IntegratorIF *createIntegrator() { return integrator = new Integrator<real>(N); }
	
	virtual real u0(real x) { return .25*exp(-(x-.5)*(x-.5)/.01); }
	virtual real t0() { return 0.; }
	virtual real tn() { return 1.; }
	virtual real dt() { return .01; }
	virtual real xmin() { return 0.; }
	virtual real xmax() { return 1.; }
	
	/*
	du/dt = d/dx[1/2 u^2]
	(u(x,d+dt) - u(x,t)) / dt = (u(x+dx,t)^2 - u(x-dx,t)^2) / (4 * dx)
	u(x,d+dt) =  u(x,t) + dt/(4*dx) * (u(x+dx,t)^2 - u(x-dx,t)^2)
	*/
	virtual void integrateForwardTimeCenteredSpace(real *un, real *u, real dt) {
		real mu = dt / dx;
		for (int i = 1; i < N-1; i++) {
			un[i] = u[i] + .25 * mu * (u[i+1]*u[i+1] - u[i-1]*u[i-1]);
		}
		un[0] = un[1];
		un[N-1] = un[N-2];
	}

	virtual void integrateLax(real *un, real *u, real dt) {}
	virtual void integrateLaxWendroff(real *un, real *u, real dt) {}
	virtual void integrateBackwardTimeCenteredSpace(real *un, real *u, real dt) {}
	virtual void integrateCrankNicolson(real *un, real *u, real dt) {}
	virtual void integrateTarget(real *un, real *u, real dt) {}
};

/*
TODO: some nonlinear 1D problems, so you can't just tridiagonal-solve everything
*/

const int DIM = 100;

typedef map<string, Problem::integrateFunc> IntegrateFuncMap;
IntegrateFuncMap integrateFuncs;

struct Test {
	Problem *problem;
	
	Test() {
		//problem = new HeatDiffuseProblem();
		//problem = new PropagationProblem();
		problem = new AdvectionProblem();

		problem->init(DIM);
	}
	
	void run() {
		integrateFuncs["ftcs"] = &Problem::integrateForwardTimeCenteredSpace;
		integrateFuncs["lax"] = &Problem::integrateLax;
		integrateFuncs["lax-wendroff"] = &Problem::integrateLaxWendroff;
		integrateFuncs["btcs"] = &Problem::integrateBackwardTimeCenteredSpace;
		integrateFuncs["crank-nicolson"] = &Problem::integrateCrankNicolson;
		integrateFuncs["target"] = &Problem::integrateTarget;
	
		for (IntegrateFuncMap::iterator i = integrateFuncs.begin(); i != integrateFuncs.end(); ++i) {
			IntegratorIF *integrator = problem->createIntegrator();
			integrator->run(problem, i->first, i->second);
		}
	}
};

int main() {
	Test test;
	test.run();
}
